/*
 * Copyright (c) 2016 Intel Corporation.
 * Copyright (c) 2019-2020 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <zephyr/kernel.h>
#include <zephyr/logging/log.h>
#include <zephyr/usb/usb_device.h>
#include <zephyr/fs/fs.h>
#include <stdio.h>
#include <zephyr/drivers/gpio.h>
#include <zephyr/dfu/flash_img.h>
#include <zephyr/storage/stream_flash.h>
#include <zephyr/dfu/mcuboot.h>
#include <zephyr/sys/reboot.h>

LOG_MODULE_REGISTER(main);


/*
 * Get button configuration from the devicetree sw0 alias. This is mandatory.
 */
#define SW0_NODE	DT_ALIAS(sw0)
#define SW1_NODE	DT_ALIAS(sw1)
#if !DT_NODE_HAS_STATUS(SW0_NODE, okay)
#error "Unsupported board: sw0 devicetree alias is not defined"
#endif
static const struct gpio_dt_spec button1 = GPIO_DT_SPEC_GET_OR(SW0_NODE, gpios,
							      {0});
static const struct gpio_dt_spec button2 = GPIO_DT_SPEC_GET_OR(SW1_NODE, gpios,
							      {0});
static struct gpio_callback button_cb_data;

/*
 * The led0 devicetree alias is optional. If present, we'll use it
 * to turn on the LED whenever the button is pressed.
 */
static struct gpio_dt_spec led = GPIO_DT_SPEC_GET_OR(DT_ALIAS(led0), gpios,
						     {0});


#if CONFIG_DISK_DRIVER_FLASH
#include <zephyr/storage/flash_map.h>
#endif

#if CONFIG_FAT_FILESYSTEM_ELM
#include <ff.h>
#endif

#if CONFIG_FILE_SYSTEM_LITTLEFS
#include <zephyr/fs/littlefs.h>
FS_LITTLEFS_DECLARE_DEFAULT_CONFIG(littlefs_storage);
#endif

#define STORAGE_PARTITION		fatfs_storage //storage_partition//external_flash_storage//littlefs_storage
#define STORAGE_PARTITION_ID		FIXED_PARTITION_ID(FATFS_STORAGE)


static struct fs_mount_t fs_mnt;

static int setup_flash(struct fs_mount_t *mnt)
{
	int rc = 0;
#if CONFIG_DISK_DRIVER_FLASH
	unsigned int id;
	const struct flash_area *pfa;

	mnt->storage_dev = (void *)STORAGE_PARTITION_ID;
	id = STORAGE_PARTITION_ID;

	rc = flash_area_open(id, &pfa);
	printk("Area %u at 0x%x on %s for %u bytes\n",
	       id, (unsigned int)pfa->fa_off, pfa->fa_dev->name,
	       (unsigned int)pfa->fa_size);

	if (rc < 0 && IS_ENABLED(CONFIG_APP_WIPE_STORAGE)) {
		printk("Erasing flash area ... ");
		rc = flash_area_erase(pfa, 0, pfa->fa_size);
		printk("%d\n", rc);
	}

	if (rc < 0) {
		flash_area_close(pfa);
	}
#endif
	return rc;
}

static int mount_app_fs(struct fs_mount_t *mnt)
{
	int rc;

#if CONFIG_FAT_FILESYSTEM_ELM
	static FATFS fat_fs;

	mnt->type = FS_FATFS;
	mnt->fs_data = &fat_fs;
	if (IS_ENABLED(CONFIG_DISK_DRIVER_RAM)) {
		mnt->mnt_point = "/RAM:";
	} else if (IS_ENABLED(CONFIG_DISK_DRIVER_SDMMC)) {
		mnt->mnt_point = "/SD:";
	} else {
		mnt->mnt_point = "/NAND:";
	}

#elif CONFIG_FILE_SYSTEM_LITTLEFS
	mnt->type = FS_LITTLEFS;
	mnt->mnt_point = "/lfs";
	mnt->fs_data = &storage;
#endif
	rc = fs_mount(mnt);

	return rc;
}

static void setup_disk(void)
{
	struct fs_mount_t *mp = &fs_mnt;
	struct fs_dir_t dir;
	struct fs_statvfs sbuf;
	int rc;

	fs_dir_t_init(&dir);

	if (IS_ENABLED(CONFIG_DISK_DRIVER_FLASH)) {
		rc = setup_flash(mp);
		if (rc < 0) {
			LOG_ERR("Failed to setup flash area");
			return;
		}
	}

	if (!IS_ENABLED(CONFIG_FILE_SYSTEM_LITTLEFS) &&
	    !IS_ENABLED(CONFIG_FAT_FILESYSTEM_ELM)) {
		LOG_INF("No file system selected");
		return;
	}

	rc = mount_app_fs(mp);
	if (rc < 0) {
		LOG_ERR("Failed to mount filesystem");
		return;
	}

	/* Allow log messages to flush to avoid interleaved output */
	k_sleep(K_MSEC(500));

	printk("Mount %s: %d\n", fs_mnt.mnt_point, rc);

	rc = fs_statvfs(mp->mnt_point, &sbuf);
	if (rc < 0) {
		LOG_ERR("FAIL: statvfs: %d\n", rc);
		return;
	}

	printk("%s: bsize = %lu ; frsize = %lu ;"
	       " blocks = %lu ; bfree = %lu\n",
	       mp->mnt_point,
	       sbuf.f_bsize, sbuf.f_frsize,
	       sbuf.f_blocks, sbuf.f_bfree);

	rc = fs_opendir(&dir, mp->mnt_point);
	printk("%s opendir: %d\n", mp->mnt_point, rc);

	if (rc < 0) {
		LOG_ERR("Failed to open directory");
	}

	while (rc >= 0) {
		struct fs_dirent ent = { 0 };

		rc = fs_readdir(&dir, &ent);
		if (rc < 0) {
			LOG_ERR("Failed to read directory entries");
			break;
		}
		if (ent.name[0] == 0) {
			printk("End of files\n");
			break;
		}
		printk("  %c %u %s\n",
		       (ent.type == FS_DIR_ENTRY_FILE) ? 'F' : 'D',
		       ent.size,
		       ent.name);
	}

	(void)fs_closedir(&dir);

	return;
}

static int firmware_update()
{
	uint8_t boot_count = 0;
	struct fs_file_t file;
	struct fs_dirent ent;
	int rc, ret;

	char fname[100];
	snprintf(fname, sizeof(fname), "%s/app_update.bin", fs_mnt.mnt_point); //application update
	fs_file_t_init(&file);
	rc = fs_open(&file, fname, FS_O_READ);
	if (rc < 0) {
		LOG_ERR("FAIL: open %s: %d", fname, rc);
		snprintf(fname, sizeof(fname), "%s/signed_by_mcuboot_and_b0_s1_image_update.bin", fs_mnt.mnt_point); //mcuboot slot S1 update
		rc = fs_open(&file, fname, FS_O_READ);
		if (rc < 0) {
			LOG_ERR("FAIL: open %s: %d", fname, rc);
			snprintf(fname, sizeof(fname), "%s/dfu_mcuboot.zip", fs_mnt.mnt_point); //both S0 and S1 binary, device select firmware
			rc = fs_open(&file, fname, FS_O_READ);
			if (rc < 0) {
				LOG_ERR("FAIL: open %s: %d", fname, rc);
				return rc;
			}
		}
	}
	rc = fs_stat(fname,&ent);

	uint8_t recv_buf[CONFIG_IMG_BLOCK_BUF_SIZE];
	uint64_t bytesRemaining,filesize;
	int fileResult;
	struct flash_img_context flash_img;
	int err = flash_img_init(&flash_img);
	if (err < 0) {
		LOG_WRN("FAIL: flash_img_init : %d", err);
		goto out;
	}

	
	bytesRemaining = ent.size; //filesize
	LOG_WRN("dfu start=%ld",bytesRemaining);
	while(bytesRemaining > 0)
	{
		int readAmount = MIN(bytesRemaining, CONFIG_IMG_BLOCK_BUF_SIZE);

		fileResult = fs_read(&file, recv_buf, readAmount);
		if(fileResult != readAmount){
			LOG_WRN("FAIL: fileResult:%d, readAmount:%d", fileResult, readAmount);
		}       
		err = flash_img_buffered_write(&flash_img, (uint8_t *) recv_buf, fileResult, false);
		if (err < 0) {
			LOG_WRN("FAIL: flash_img_buffered_write : %d", err);
			goto out;
		}

		bytesRemaining -= fileResult;
	}
	LOG_WRN("dfu end=%ld",bytesRemaining);
	// Write with 0 length to flush the write operation to flash. 
	err = flash_img_buffered_write(&flash_img,(uint8_t *)recv_buf, 0, true);
	if (err < 0) {
			LOG_WRN("FAIL: flash_img_buffered_write0 : %d", err);
			goto out;
	}

	err = boot_request_upgrade(BOOT_UPGRADE_PERMANENT); 
	if (err < 0) {
			LOG_ERR("FAIL: boot_request_upgrade : %d", err);
			goto out;
	}

	k_sleep(K_MSEC(1000));
	// Reboot system to apply the firmware.
	sys_reboot(SYS_REBOOT_COLD); 

 out:
	ret = fs_close(&file);
	if (ret < 0) {
		LOG_ERR("FAIL: close %s: %d", fname, ret);
		return ret;
	}

	return (rc < 0 ? rc : 0);
}

void button_pressed(const struct device *dev, struct gpio_callback *cb,
		    uint32_t pins)
{
	printk("Button pressed at %" PRIu32 "\n", k_cycle_get_32());
}

void main(void)
{
	int ret;
	LOG_WRN("Image build time: " __DATE__ " " __TIME__);

	if (!device_is_ready(button1.port)) {
		LOG_INF("Error: button device %s is not ready\n",
		       button1.port->name);
		return;
	}

	ret = gpio_pin_configure_dt(&button1, GPIO_INPUT);
	if (ret != 0) {
		LOG_INF("Error %d: failed to configure %s pin %d\n",
		       ret, button1.port->name, button1.pin);
		return;
	}

	ret = gpio_pin_configure_dt(&button2, GPIO_INPUT);
	if (ret != 0) {
		LOG_INF("Error %d: failed to configure %s pin %d\n",
		       ret, button2.port->name, button2.pin);
		return;
	}

	ret = gpio_pin_interrupt_configure_dt(&button1,
					      GPIO_INT_EDGE_TO_ACTIVE);
	if (ret != 0) {
		LOG_INF("Error %d: failed to configure interrupt on %s pin %d\n",
			ret, button1.port->name, button1.pin);
		return;
	}

	gpio_init_callback(&button_cb_data, button_pressed, BIT(button1.pin));
	gpio_add_callback(button1.port, &button_cb_data);
	LOG_INF("Set up button at %s pin %d\n", button1.port->name, button1.pin);

	bool image_ok = boot_is_img_confirmed();
	LOG_INF("Image is%s confirmed OK", image_ok ? "" : " not");
	if (!image_ok) {
		ret = boot_write_img_confirmed();
		if (ret < 0) {
			LOG_ERR("Couldn't confirm this image: %d", ret);
			return;
		}
		LOG_WRN("Marked image as OK");
		ret = boot_erase_img_bank(FIXED_PARTITION_ID(slot1_partition));
		if (ret) {
			LOG_ERR("Failed to erase second slot: %d", ret);
			return;
		}
	}

	setup_disk();
	
	while (1) {
		if (gpio_pin_get_dt(&button1)) {
			firmware_update();
		}
		if (gpio_pin_get_dt(&button2)) {
			LOG_WRN("button2 preesed...");
			break;
		}
		k_sleep(K_MSEC(100));
	}
	
	LOG_WRN("The device is put in USB mass storage mode,please reset to restart the program.\n");
	ret = usb_enable(NULL);
	if (ret != 0) {
		LOG_ERR("Failed to enable USB");
		return;
	}
}


